@@ -26,15 +26,17 @@
 #include "screen-cast.h"
 #include "remote-desktop.h"
 #include "request.h"
+#include "permissions.h"
 #include "pipewire.h"
 #include "xdp-dbus.h"
 #include "xdp-impl-dbus.h"
 #include "xdp-utils.h"
 
-#define PERMISSION_ITEM(item_key, item_value) \
-  ((struct spa_dict_item) { \
-    .key = item_key, \
-    .value = item_value \
+#define RESTORE_DATA_TYPE "(suv)"
+#define PERMISSION_ITEM(item_id, item_permissions) \
+  ((struct pw_permission) { \
+    .id = item_id, \
+    .permissions = item_permissions \
   })
 
 typedef struct _ScreenCast ScreenCast;
@@ -54,6 +56,9 @@ static XdpImplScreenCast *impl;
 static int impl_version;
 static ScreenCast *screen_cast;
 
+static GMutex transient_permissions_lock;
+static GHashTable *transient_permissions;
+
 static unsigned int available_cursor_modes = 0;
 
 GType screen_cast_get_type (void);
@@ -72,6 +77,13 @@ G_DEFINE_TYPE_WITH_CODE (ScreenCast, screen_cast, XDP_TYPE_SCREEN_CAST_SKELETON,
                          G_IMPLEMENT_INTERFACE (XDP_TYPE_SCREEN_CAST,
                                                 screen_cast_iface_init))
 
+typedef enum _PersistMode
+{
+  PERSIST_MODE_NONE = 0,
+  PERSIST_MODE_TRANSIENT = 1,
+  PERSIST_MODE_PERSISTENT = 2,
+} PersistMode;
+
 typedef enum _ScreenCastSessionState
 {
   SCREEN_CAST_SESSION_STATE_INIT,
@@ -89,6 +101,9 @@ typedef struct _ScreenCastSession
   ScreenCastSessionState state;
 
   GList *streams;
+  char *restore_token;
+  PersistMode persist_mode;
+  GVariant *restore_data;
 } ScreenCastSession;
 
 typedef struct _ScreenCastSessionClass
@@ -100,6 +115,145 @@ GType screen_cast_session_get_type (void);
 
 G_DEFINE_TYPE (ScreenCastSession, screen_cast_session, session_get_type ())
 
+static void
+set_persistent_permissions (const char *app_id,
+                            const char *restore_token,
+                            GVariant *restore_data)
+{
+  g_autoptr(GError) error = NULL;
+
+  set_permission_sync (app_id, "screencast", restore_token, PERMISSION_YES);
+
+  if (!xdp_impl_permission_store_call_set_value_sync (get_permission_store (),
+                                                      "screencast",
+                                                      TRUE,
+                                                      restore_token,
+                                                      g_variant_new_variant (restore_data),
+                                                      NULL,
+                                                      &error))
+    {
+      g_dbus_error_strip_remote_error (error);
+      g_warning ("Error setting permission store value: %s", error->message);
+    }
+}
+
+static GVariant *
+get_persistent_permissions (const char *app_id,
+                            const char *restore_token)
+{
+  g_autoptr(GVariant) perms = NULL;
+  g_autoptr(GVariant) data = NULL;
+  g_autoptr(GError) error = NULL;
+  const char **permissions;
+
+  if (!xdp_impl_permission_store_call_lookup_sync (get_permission_store (),
+                                                   "screencast",
+                                                   restore_token,
+                                                   &perms,
+                                                   &data,
+                                                   NULL,
+                                                   &error))
+    {
+      return NULL;
+    }
+
+  if (!perms || !g_variant_lookup (perms, app_id, "^a&s", &permissions))
+    return NULL;
+
+  if (!data)
+    return NULL;
+
+  return g_variant_get_child_value (data, 0);
+}
+
+void
+delete_persistent_permissions (const char *app_id,
+                               const char *restore_token)
+{
+
+  g_autoptr(GError) error = NULL;
+
+  if (!xdp_impl_permission_store_call_delete_sync (get_permission_store (),
+                                                   "screencast",
+                                                   restore_token,
+                                                   NULL,
+                                                   &error))
+    {
+      g_dbus_error_strip_remote_error (error);
+      g_warning ("Error deleting permission: %s", error->message);
+    }
+}
+
+static void
+set_transient_permissions (const char *sender,
+                           const char *restore_token,
+                           GVariant *restore_data)
+{
+  g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&transient_permissions_lock);
+
+  if (!transient_permissions)
+    {
+      transient_permissions = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                     g_free, (GDestroyNotify)g_variant_unref);
+    }
+
+  g_hash_table_insert (transient_permissions,
+                       g_strdup_printf ("%s/%s", sender, restore_token),
+                       g_variant_ref (restore_data));
+}
+
+static GVariant *
+get_transient_permissions (const char *sender,
+                           const char *restore_token)
+{
+  g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&transient_permissions_lock);
+  g_autofree char *id = NULL;
+  GVariant *permissions;
+
+  if (!transient_permissions)
+    return NULL;
+
+  id = g_strdup_printf ("%s/%s", sender, restore_token);
+  permissions = g_hash_table_lookup (transient_permissions, id);
+  return permissions ? g_variant_ref (permissions) : NULL;
+}
+
+static void
+delete_transient_permissions (const char *sender,
+                              const char *restore_token)
+{
+  g_autoptr(GMutexLocker) locker = g_mutex_locker_new (&transient_permissions_lock);
+  g_autofree char *id = NULL;
+
+  if (!transient_permissions)
+    return;
+
+  id = g_strdup_printf ("%s/%s", sender, restore_token);
+  g_hash_table_remove (transient_permissions, id);
+}
+
+void
+screen_cast_remove_transient_permissions_for_sender (const char *sender)
+{
+  g_autoptr(GMutexLocker) locker = NULL;
+  GHashTableIter iter;
+  const char *key;
+
+  locker = g_mutex_locker_new (&transient_permissions_lock);
+
+  if (!transient_permissions)
+    return;
+
+  g_hash_table_iter_init (&iter, transient_permissions);
+  while (g_hash_table_iter_next (&iter, (gpointer *) &key, NULL))
+    {
+      g_auto(GStrv) split = g_strsplit (key, "/", 2);
+
+      if (split && split[0] && g_strcmp0 (split[0], sender) == 0)
+        g_hash_table_iter_remove (&iter);
+    }
+}
+
 static gboolean
 is_screen_cast_session (Session *session)
 {
@@ -163,6 +317,7 @@ create_session_done (GObject *source_object,
                                                         res,
                                                         &error))
     {
+      g_dbus_error_strip_remote_error (error);
       g_warning ("A backend call failed: %s", error->message);
       should_close_session = TRUE;
       goto out;
@@ -229,7 +384,7 @@ handle_create_session (XdpScreenCast *object,
   if (!impl_request)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   request_set_impl_request (request, impl_request);
@@ -239,7 +394,7 @@ handle_create_session (XdpScreenCast *object,
   if (!session)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   g_variant_builder_init (&options_builder, G_VARIANT_TYPE_VARDICT);
@@ -261,7 +416,7 @@ handle_create_session (XdpScreenCast *object,
 
   xdp_screen_cast_complete_create_session (object, invocation, request->id);
 
-  return TRUE;
+  return G_DBUS_METHOD_INVOCATION_HANDLED;
 }
 
 static void
@@ -287,7 +442,10 @@ select_sources_done (GObject *source_object,
                                                         &results,
                                                         res,
                                                         &error))
-    g_warning ("A backend call failed: %s", error->message);
+    {
+      g_dbus_error_strip_remote_error (error);
+      g_warning ("A backend call failed: %s", error->message);
+    }
 
   should_close_session = !request->exported || response != 0;
 
@@ -338,10 +496,10 @@ validate_device_types (const char *key,
 {
   guint32 types = g_variant_get_uint32 (value);
 
-  if ((types & ~(1 | 2)) != 0)
+  if ((types & ~(1 | 2 | 4)) != 0)
     {
       g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
-                   "Unsupported device type: %x", types & ~(1 | 2));
+                   "Unsupported device type: %x", types & ~(1 | 2 | 4));
       return FALSE;
     } 
 
@@ -373,12 +531,148 @@ validate_cursor_mode (const char *key,
   return TRUE;
 }
 
+static gboolean
+validate_restore_token (const char *key,
+                        GVariant *value,
+                        GVariant *options,
+                        GError **error)
+{
+  const char *restore_token = g_variant_get_string (value, NULL);
+
+  if (!g_uuid_string_is_valid (restore_token))
+    {
+      g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
+                   "Restore token is not a valid UUID string");
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
+static gboolean
+validate_persist_mode (const char *key,
+                       GVariant *value,
+                       GVariant *options,
+                       GError **error)
+{
+  uint32_t mode = g_variant_get_uint32 (value);
+
+  if (mode > PERSIST_MODE_PERSISTENT)
+    {
+      g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
+                   "Invalid persist mode %x", mode);
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
 static XdpOptionKey screen_cast_select_sources_options[] = {
   { "types", G_VARIANT_TYPE_UINT32, validate_device_types },
   { "multiple", G_VARIANT_TYPE_BOOLEAN, NULL },
   { "cursor_mode", G_VARIANT_TYPE_UINT32, validate_cursor_mode },
+  { "restore_token", G_VARIANT_TYPE_STRING, validate_restore_token },
+  { "persist_mode", G_VARIANT_TYPE_UINT32, validate_persist_mode },
 };
 
+static gboolean
+replace_restore_token_with_data (ScreenCastSession *screen_cast_session,
+                                 GVariant **in_out_options,
+                                 GError **error)
+{
+  Session *session = (Session *)screen_cast_session;
+  GVariantBuilder options_builder;
+  g_autoptr(GVariant) options = NULL;
+  gsize i;
+
+  options = *in_out_options;
+
+  if (!g_variant_lookup (options, "persist_mode", "u", &screen_cast_session->persist_mode))
+    screen_cast_session->persist_mode = PERSIST_MODE_NONE;
+
+  if (is_remote_desktop_session (session) && screen_cast_session->persist_mode != PERSIST_MODE_NONE)
+    {
+      g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
+                   "Remote desktop sessions cannot persist");
+      return FALSE;
+    }
+
+  g_variant_builder_init (&options_builder, G_VARIANT_TYPE_VARDICT);
+  for (i = 0; i < G_N_ELEMENTS (screen_cast_select_sources_options); i++)
+    {
+      g_autoptr(GVariant) value = NULL;
+
+      value = g_variant_lookup_value (options,
+                                      screen_cast_select_sources_options[i].key,
+                                      screen_cast_select_sources_options[i].type);
+
+      if (!value)
+        continue;
+
+      if (g_strcmp0 (screen_cast_select_sources_options[i].key, "restore_token") == 0)
+        {
+          g_autoptr(GVariant) restore_data = NULL;
+          g_autofree char *restore_token = NULL;
+
+          if (is_remote_desktop_session (session))
+            {
+              g_set_error (error, XDG_DESKTOP_PORTAL_ERROR, XDG_DESKTOP_PORTAL_ERROR_INVALID_ARGUMENT,
+                           "Remote desktop sessions cannot be restored");
+              return FALSE;
+            }
+
+          restore_token = g_variant_dup_string (value, NULL);
+
+          /* Lookup permissions in memory first, and fallback to the permission
+           * store if not found. Immediately delete them now as a safety measure,
+           * since they'll be stored again when the session is closed.
+           *
+           * Notice that transient mode uses the sender name, whereas persistent
+           * mode uses the app id.
+           */
+          restore_data = get_transient_permissions (session->sender, restore_token);
+          if (restore_data)
+            {
+              delete_transient_permissions (session->sender, restore_token);
+            }
+          else
+            {
+              restore_data = get_persistent_permissions (session->app_id, restore_token);
+              if (restore_data)
+                delete_persistent_permissions (session->app_id, restore_token);
+            }
+
+          if (!restore_data)
+            continue;
+
+          if (!g_variant_check_format_string (restore_data, RESTORE_DATA_TYPE, FALSE))
+            {
+              g_warning ("Restore data was stored with the wrong format, ignoring");
+              continue;
+            }
+
+          /* Only store the restore token after checking if it exists, otherwise
+           * apps can pass random UUIDs and yet predict what the next token will
+           * be.
+           */
+          screen_cast_session->restore_token = g_steal_pointer (&restore_token);
+
+          g_debug ("Replacing 'restore_token' with portal-specific data");
+          g_variant_builder_add (&options_builder, "{sv}", "restore_data", restore_data);
+        }
+      else
+        {
+          g_variant_builder_add (&options_builder, "{sv}",
+                                 screen_cast_select_sources_options[i].key,
+                                 g_steal_pointer (&value));
+        }
+    }
+
+  *in_out_options = g_variant_builder_end (&options_builder);
+
+  return TRUE;
+}
+
 static gboolean
 handle_select_sources (XdpScreenCast *object,
                        GDBusMethodInvocation *invocation,
@@ -387,9 +681,11 @@ handle_select_sources (XdpScreenCast *object,
 {
   Request *request = request_from_invocation (invocation);
   Session *session;
+  ScreenCastSession *screen_cast_session;
   g_autoptr(GError) error = NULL;
   g_autoptr(XdpImplRequest) impl_request = NULL;
   GVariantBuilder options_builder;
+  GVariant *options;
 
   REQUEST_AUTOLOCK (request);
 
@@ -400,7 +696,7 @@ handle_select_sources (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_ACCESS_DENIED,
                                              "Invalid session");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   SESSION_AUTOLOCK_UNREF (session);
@@ -419,20 +715,20 @@ handle_select_sources (XdpScreenCast *object,
                                                  G_DBUS_ERROR,
                                                  G_DBUS_ERROR_FAILED,
                                                  "Sources already selected");
-          return TRUE;
+          return G_DBUS_METHOD_INVOCATION_HANDLED;
         case SCREEN_CAST_SESSION_STATE_STARTING:
         case SCREEN_CAST_SESSION_STATE_STARTED:
           g_dbus_method_invocation_return_error (invocation,
                                                  G_DBUS_ERROR,
                                                  G_DBUS_ERROR_FAILED,
                                                  "Can only select sources before starting");
-          return TRUE;
+          return G_DBUS_METHOD_INVOCATION_HANDLED;
         case SCREEN_CAST_SESSION_STATE_CLOSED:
           g_dbus_method_invocation_return_error (invocation,
                                                  G_DBUS_ERROR,
                                                  G_DBUS_ERROR_FAILED,
                                                  "Invalid session");
-          return TRUE;
+          return G_DBUS_METHOD_INVOCATION_HANDLED;
         }
     }
   else if (is_remote_desktop_session (session))
@@ -446,7 +742,7 @@ handle_select_sources (XdpScreenCast *object,
                                                  G_DBUS_ERROR,
                                                  G_DBUS_ERROR_FAILED,
                                                  "Invalid state");
-          return TRUE;
+          return G_DBUS_METHOD_INVOCATION_HANDLED;
         }
     }
   else
@@ -455,7 +751,7 @@ handle_select_sources (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_FAILED,
                                              "Invalid session");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   impl_request =
@@ -467,7 +763,7 @@ handle_select_sources (XdpScreenCast *object,
   if (!impl_request)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   request_set_impl_request (request, impl_request);
@@ -480,7 +776,20 @@ handle_select_sources (XdpScreenCast *object,
                            &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
+    }
+
+  options = g_variant_builder_end (&options_builder);
+
+  /* If 'restore_token' is passed, lookup the corresponding data in the
+   * permission store and / or the GHashTable with transient permissions.
+   * Portal implementations do not have access to the restore token.
+   */
+  screen_cast_session = (ScreenCastSession *) session;
+  if (!replace_restore_token_with_data (screen_cast_session, &options, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   g_object_set_qdata_full (G_OBJECT (request),
@@ -501,14 +810,14 @@ handle_select_sources (XdpScreenCast *object,
                                             request->id,
                                             arg_session_handle,
                                             xdp_app_info_get_id (request->app_info),
-                                            g_variant_builder_end (&options_builder),
+                                            options,
                                             NULL,
                                             select_sources_done,
                                             g_object_ref (request));
 
   xdp_screen_cast_complete_select_sources (object, invocation, request->id);
 
-  return TRUE;
+  return G_DBUS_METHOD_INVOCATION_HANDLED;
 }
 
 uint32_t
@@ -517,42 +826,9 @@ screen_cast_stream_get_pipewire_node_id (ScreenCastStream *stream)
   return stream->id;
 }
 
-static void
-append_parent_permissions (PipeWireRemote *remote,
-                           GArray *permission_items,
-                           GList **string_stash,
-                           PipeWireGlobal *global,
-                           const char *permission)
-{
-  PipeWireGlobal *parent;
-  char *parent_permission_value;
-
-  if (global->parent_id == 0)
-    return;
-
-  parent = g_hash_table_lookup (remote->globals, GINT_TO_POINTER (global->parent_id));
-
-  if (parent->permission_set)
-    return;
-  parent->permission_set = TRUE;
-
-  append_parent_permissions (remote, permission_items, string_stash,
-                             parent, permission);
-
-  parent_permission_value = g_strdup_printf ("%u:%s",
-                                             global->parent_id,
-                                             permission);
-  *string_stash = g_list_prepend (*string_stash, parent_permission_value);
-
-  g_array_append_val (permission_items,
-                      PERMISSION_ITEM (PW_CORE_PROXY_PERMISSIONS_GLOBAL,
-                                       parent_permission_value));
-}
-
 static void
 append_stream_permissions (PipeWireRemote *remote,
                            GArray *permission_items,
-                           GList **string_stash,
                            GList *streams)
 {
   GList *l;
@@ -561,21 +837,10 @@ append_stream_permissions (PipeWireRemote *remote,
     {
       ScreenCastStream *stream = l->data;
       uint32_t stream_id;
-      PipeWireGlobal *stream_global;
-      char *stream_permission_value;
 
       stream_id = screen_cast_stream_get_pipewire_node_id (stream);
-      stream_global = g_hash_table_lookup (remote->globals,
-                                           GINT_TO_POINTER (stream_id));
-
-      append_parent_permissions (remote, permission_items, string_stash,
-                                 stream_global, "r--");
-
-      stream_permission_value = g_strdup_printf ("%u:rwx", stream_id);
-      *string_stash = g_list_prepend (*string_stash, stream_permission_value);
       g_array_append_val (permission_items,
-                          PERMISSION_ITEM (PW_CORE_PROXY_PERMISSIONS_GLOBAL,
-                                           stream_permission_value));
+                          PERMISSION_ITEM (stream_id, PW_PERM_RWX));
     }
 }
 
@@ -587,10 +852,6 @@ open_pipewire_screen_cast_remote (const char *app_id,
   struct pw_properties *pipewire_properties;
   PipeWireRemote *remote;
   g_autoptr(GArray) permission_items = NULL;
-  char *node_factory_permission_string;
-  GList *string_stash = NULL;
-  struct spa_dict *permission_dict;
-  PipeWireGlobal *node_global;
 
   pipewire_properties = pw_properties_new ("pipewire.access.portal.app_id", app_id,
                                            "pipewire.access.portal.media_roles", "",
@@ -601,50 +862,33 @@ open_pipewire_screen_cast_remote (const char *app_id,
   if (!remote)
     return FALSE;
 
-  permission_items = g_array_new (FALSE, TRUE, sizeof (struct spa_dict_item));
-
-  /*
-   * Hide all existing and future nodes (except the ones we explicitly list below.
-   */
-  g_array_append_val (permission_items,
-                      PERMISSION_ITEM (PW_CORE_PROXY_PERMISSIONS_EXISTING,
-                                       "---"));
-  g_array_append_val (permission_items,
-                      PERMISSION_ITEM (PW_CORE_PROXY_PERMISSIONS_DEFAULT,
-                                       "---"));
+  permission_items = g_array_new (FALSE, TRUE, sizeof (struct pw_permission));
 
   /*
    * PipeWire:Interface:Core
    * Needs rwx to be able create the sink node using the create-object method
    */
   g_array_append_val (permission_items,
-                      PERMISSION_ITEM (PW_CORE_PROXY_PERMISSIONS_GLOBAL,
-                                       "0:rwx"));
+                      PERMISSION_ITEM (PW_ID_CORE, PW_PERM_RWX));
 
   /*
    * PipeWire:Interface:NodeFactory
    * Needs r-- so it can be passed to create-object when creating the sink node.
    */
-  node_factory_permission_string = g_strdup_printf ("%d:r--",
-                                                    remote->node_factory_id);
-  string_stash = g_list_prepend (string_stash, node_factory_permission_string);
   g_array_append_val (permission_items,
-                      PERMISSION_ITEM (PW_CORE_PROXY_PERMISSIONS_GLOBAL,
-                                       node_factory_permission_string));
-  node_global = g_hash_table_lookup (remote->globals,
-                                     GINT_TO_POINTER (remote->node_factory_id));
-  append_parent_permissions (remote, permission_items, &string_stash,
-                             node_global, "r--");
+                      PERMISSION_ITEM (remote->node_factory_id, PW_PERM_R));
 
-  append_stream_permissions (remote, permission_items, &string_stash, streams);
+  append_stream_permissions (remote, permission_items, streams);
 
-  permission_dict =
-    &SPA_DICT_INIT ((struct spa_dict_item *) permission_items->data,
-                    permission_items->len);
-  pw_core_proxy_permissions (pw_remote_get_core_proxy (remote->remote),
-                             permission_dict);
+  /*
+   * Hide all existing and future nodes (except the ones we explicitly list above).
+   */
+  g_array_append_val (permission_items,
+                      PERMISSION_ITEM (PW_ID_ANY, 0));
 
-  g_list_free_full (string_stash, g_free);
+  pw_client_update_permissions (pw_core_get_client(remote->core),
+                                permission_items->len,
+                                (const struct pw_permission *)permission_items->data);
 
   pipewire_remote_roundtrip (remote);
 
@@ -689,12 +933,121 @@ collect_screen_cast_stream_data (GVariantIter *streams_iter)
   return streams;
 }
 
+static void
+generate_and_save_restore_token (ScreenCastSession *screen_cast_session)
+{
+  Session *session = (Session *)screen_cast_session;
+
+  if (!screen_cast_session->restore_data)
+    {
+      if (screen_cast_session->restore_token)
+        {
+          delete_persistent_permissions (session->app_id, screen_cast_session->restore_token);
+          delete_transient_permissions (session->sender, screen_cast_session->restore_token);
+        }
+
+      g_clear_pointer (&screen_cast_session->restore_token, g_free);
+      return;
+    }
+
+  switch (screen_cast_session->persist_mode)
+    {
+    case PERSIST_MODE_NONE:
+      if (screen_cast_session->restore_token)
+        {
+          delete_persistent_permissions (session->app_id, screen_cast_session->restore_token);
+          delete_transient_permissions (session->sender, screen_cast_session->restore_token);
+        }
+
+      g_clear_pointer (&screen_cast_session->restore_token, g_free);
+      g_clear_pointer (&screen_cast_session->restore_data, g_variant_unref);
+      break;
+
+    case PERSIST_MODE_TRANSIENT:
+      if (screen_cast_session->restore_token == NULL)
+        screen_cast_session->restore_token = g_uuid_string_random ();
+
+      set_transient_permissions (session->sender,
+                                 screen_cast_session->restore_token,
+                                 screen_cast_session->restore_data);
+      break;
+
+    case PERSIST_MODE_PERSISTENT:
+      if (screen_cast_session->restore_token == NULL)
+        screen_cast_session->restore_token = g_uuid_string_random ();
+
+      set_persistent_permissions (session->app_id,
+                                  screen_cast_session->restore_token,
+                                  screen_cast_session->restore_data);
+
+      break;
+    }
+}
+
+static void
+replace_restore_data_by_token (ScreenCastSession *screen_cast_session,
+                               GVariant **in_out_results)
+{
+  g_autoptr(GVariant) results = *in_out_results;
+  GVariantBuilder results_builder;
+  GVariantIter iter;
+  const char *key;
+  GVariant *value;
+  gboolean found_restore_data = FALSE;
+
+  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
+
+  g_variant_iter_init (&iter, results);
+  while (g_variant_iter_next (&iter, "{&sv}", &key, &value))
+    {
+      if (g_strcmp0 (key, "restore_data") == 0)
+        {
+          if (g_variant_check_format_string (value, RESTORE_DATA_TYPE, FALSE))
+            {
+              screen_cast_session->restore_data = g_variant_ref_sink (value);
+              found_restore_data = TRUE;
+            }
+          else
+            {
+              g_warning ("Received restore data in invalid variant format ('%s'; expected '%s')",
+                         g_variant_get_type_string (value),
+                         RESTORE_DATA_TYPE);
+            }
+        }
+      else if (g_strcmp0 (key, "persist_mode") == 0)
+        {
+          screen_cast_session->persist_mode = MIN (screen_cast_session->persist_mode,
+                                                   g_variant_get_uint32 (value));
+        }
+      else
+        {
+          g_variant_builder_add (&results_builder, "{sv}", key, value);
+        }
+    }
+
+  if (found_restore_data)
+    {
+      g_debug ("Replacing restore data received from portal impl with a token");
+
+      generate_and_save_restore_token (screen_cast_session);
+      g_variant_builder_add (&results_builder, "{sv}", "restore_token",
+                             g_variant_new_string (screen_cast_session->restore_token));
+    }
+  else
+    {
+      screen_cast_session->persist_mode = PERSIST_MODE_NONE;
+    }
+
+  *in_out_results = g_variant_builder_end (&results_builder);
+}
+
 static gboolean
 process_results (ScreenCastSession *screen_cast_session,
-                 GVariant *results,
+                 GVariant **in_out_results,
                  GError **error)
 {
   g_autoptr(GVariantIter) streams_iter = NULL;
+  GVariant *results = *in_out_results;
 
   if (!g_variant_lookup (results, "streams", "a(ua{sv})", &streams_iter))
     {
@@ -703,6 +1056,7 @@ process_results (ScreenCastSession *screen_cast_session,
     }
 
   screen_cast_session->streams = collect_screen_cast_stream_data (streams_iter);
+  replace_restore_data_by_token (screen_cast_session, in_out_results);
   return TRUE;
 }
 
@@ -730,7 +1084,10 @@ start_done (GObject *source_object,
                                                &results,
                                                res,
                                                &error))
-    g_warning ("A backend call failed: %s", error->message);
+    {
+      g_dbus_error_strip_remote_error (error);
+      g_warning ("A backend call failed: %s", error->message);
+    }
 
   should_close_session = !request->exported || response != 0;
 
@@ -740,7 +1097,7 @@ start_done (GObject *source_object,
     {
       if (response == 0)
         {
-          if (!process_results (screen_cast_session, results, &error))
+          if (!process_results (screen_cast_session, &results, &error))
             {
               g_warning ("Failed to process results: %s", error->message);
               g_clear_error (&error);
@@ -799,7 +1156,7 @@ handle_start (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_ACCESS_DENIED,
                                              "Invalid session");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   SESSION_AUTOLOCK_UNREF (session);
@@ -815,20 +1172,20 @@ handle_start (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_FAILED,
                                              "Sources not selected");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     case SCREEN_CAST_SESSION_STATE_STARTING:
     case SCREEN_CAST_SESSION_STATE_STARTED:
       g_dbus_method_invocation_return_error (invocation,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_FAILED,
                                              "Can only start once");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     case SCREEN_CAST_SESSION_STATE_CLOSED:
       g_dbus_method_invocation_return_error (invocation,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_FAILED,
                                              "Invalid session");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   g_object_set_data_full (G_OBJECT (request),
@@ -843,7 +1200,7 @@ handle_start (XdpScreenCast *object,
   if (!impl_request)
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   request_set_impl_request (request, impl_request);
@@ -870,7 +1227,7 @@ handle_start (XdpScreenCast *object,
 
   xdp_screen_cast_complete_start (object, invocation, request->id);
 
-  return TRUE;
+  return G_DBUS_METHOD_INVOCATION_HANDLED;
 }
 
 static gboolean
@@ -896,7 +1253,7 @@ handle_open_pipewire_remote (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_ACCESS_DENIED,
                                              "Invalid session");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   SESSION_AUTOLOCK_UNREF (session);
@@ -920,7 +1277,7 @@ handle_open_pipewire_remote (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_ACCESS_DENIED,
                                              "Invalid session");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   if (!streams)
@@ -929,7 +1286,7 @@ handle_open_pipewire_remote (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_FAILED,
                                              "No streams available");
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   remote = open_pipewire_screen_cast_remote (session->app_id, streams, &error);
@@ -939,11 +1296,11 @@ handle_open_pipewire_remote (XdpScreenCast *object,
                                              G_DBUS_ERROR,
                                              G_DBUS_ERROR_FAILED,
                                              "%s", error->message);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   out_fd_list = g_unix_fd_list_new ();
-  fd = pw_remote_steal_fd (remote->remote);
+  fd = pw_core_steal_fd (remote->core);
   fd_id = g_unix_fd_list_append (out_fd_list, fd, &error);
   close (fd);
   pipewire_remote_destroy (remote);
@@ -955,13 +1312,13 @@ handle_open_pipewire_remote (XdpScreenCast *object,
                                              G_DBUS_ERROR_ACCESS_DENIED,
                                              "Failed to append fd: %s",
                                              error->message);
-      return TRUE;
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
     }
 
   xdp_screen_cast_complete_open_pipewire_remote (object, invocation,
                                                  out_fd_list,
                                                  g_variant_new_handle (fd_id));
-  return TRUE;
+  return G_DBUS_METHOD_INVOCATION_HANDLED;
 }
 
 static void
@@ -1011,7 +1368,7 @@ on_supported_cursor_modes_changed (GObject *gobject,
 static void
 screen_cast_init (ScreenCast *screen_cast)
 {
-  xdp_screen_cast_set_version (XDP_SCREEN_CAST (screen_cast), 2);
+  xdp_screen_cast_set_version (XDP_SCREEN_CAST (screen_cast), 4);
 
   g_signal_connect (impl, "notify::supported-source-types",
                     G_CALLBACK (on_supported_source_types_changed),
@@ -1067,6 +1424,8 @@ screen_cast_session_close (Session *session)
 
   screen_cast_session->state = SCREEN_CAST_SESSION_STATE_CLOSED;
 
+  generate_and_save_restore_token (screen_cast_session);
+
   g_debug ("screen cast session owned by '%s' closed", session->sender);
 }
 
@@ -1075,6 +1434,9 @@ screen_cast_session_finalize (GObject *object)
 {
   ScreenCastSession *screen_cast_session = (ScreenCastSession *)object;
 
+  g_clear_pointer (&screen_cast_session->restore_token, g_free);
+  g_clear_pointer (&screen_cast_session->restore_data, g_variant_unref);
+
   g_list_free_full (screen_cast_session->streams,
                     (GDestroyNotify)screen_cast_stream_free);
 
